class QueueKeys:
    """
    This class handles all keys parser logic.
    """

    def __init__(self, prefix: str = 'bull'):
        self.prefix = prefix

    def getKeys(self, name: str):
        names = ["", "active", "wait", "waiting-children", "paused", "completed", "failed", "delayed",
                 "stalled", "limiter", "prioritized", "id", "stalled-check", "meta", "pc", "events", "marker"]
        keys = {}
        for name_type in names:
            keys[name_type] = self.toKey(name, name_type)

        return keys

    def toKey(self, name: str, name_type: str):
        return f"{self.getQueueQualifiedName(name)}:{name_type}"

    def getQueueQualifiedName(self, name: str):
        return f"{self.prefix}:{name}"



from bullmq.redis_connection import RedisConnection
from bullmq.types import QueueBaseOptions
from bullmq.scripts import Scripts
from bullmq.job import Job
from bullmq.queue_keys import QueueKeys
from uuid import uuid4


class MinimalQueue:
    """
    Instantiate a MinimalQueue object
    """

    def __init__(self, name: str, queue_keys, redisConnection, opts: QueueBaseOptions = {}):
        """
        Initialize a connection
        """
        self.name = name
        self.redisConnection = redisConnection
        self.client = self.redisConnection.conn
        self.opts = opts
        self.prefix = opts.get("prefix", "bull"),
        self.keys = queue_keys.getKeys(name)
        self.qualifiedName = queue_keys.getQueueQualifiedName(name)


class FlowProducer:
    """
    Instantiate a FlowProducer object
    """

    def __init__(self, redisOpts: dict | str = {}, opts: QueueBaseOptions = {}):
        """
        Initialize a connection
        """
        self.redisConnection = RedisConnection(redisOpts)
        self.client = self.redisConnection.conn
        self.opts: dict = opts
        self.prefix = opts.get("prefix", "bull")
        self.scripts = Scripts(
            self.prefix, "__default__", self.redisConnection)

    def queueFromNode( self, node:dict, queue_keys, prefix: str):
        return MinimalQueue(node.get("queueName"),queue_keys,self.redisConnection, {"prefix": prefix})

    async def addChildren(self, nodes, parent, queues_opts, pipe):
        children = []
        for node in nodes:
            job = await self.addNode(node, parent, queues_opts, pipe)
            children.append(job)
        return children

    async def addNodes(self, nodes: list[dict], pipe):
        trees = []
        for node in nodes:
            parent_opts = node.get("opts", {}).get("parent", None)
            jobs_tree = await self.addNode(node, {"parentOpts": parent_opts},None, pipe)
            trees.append(jobs_tree)

        return trees

    async def addNode(self, node: dict, parent: dict, queues_opts: dict, pipe):
        prefix = node.get("prefix", self.prefix)
        queue = self.queueFromNode(node, QueueKeys(prefix), prefix)
        queue_name = node.get("queueName")
        queue_opts = queues_opts and queues_opts.get(queue_name)

        jobs_opts = queue_opts.get('defaultJobOptions',{}) if queue_opts else {}
        jobs_opts.update(node.get("opts") or {})
        job_id = (node.get("opts") or {}).get("jobId") or uuid4().hex
        parent_opts = parent.get("parentOpts")
        
        jobs_opts.update({"parent": parent_opts})

        job = Job(
            queue=queue,
            name=node.get("name"),
            data=node.get("data"),
            opts=jobs_opts,
            job_id = job_id
            )

        node_children = node.get("children", [])

        self.scripts.resetQueueKeys(queue_name)
        if len(node_children)>0:
            parent_id = job_id
            queue_keys_parent = QueueKeys(prefix or self.opts.get("prefix", "bull"))
            wait_children_key = queue_keys_parent.toKey(queue_name, "waiting-children")

            await self.scripts.addParentJob(
                job,
                wait_children_key,
                pipe
            )

            children = await self.addChildren(
                node_children,
                {  
                    "parentOpts": {
                        "id": parent_id,
                        "queue": queue.qualifiedName
                    } 
                },
                queues_opts,
                pipe
                )
            return {"job": job, "children": children}
        else:
            await self.scripts.addJob(
                job,
                pipe
            )

            return {"job": job}

    async def add(self, flow: dict, opts: dict = {}):
        parent_opts = flow.get("opts", {}).get("parent", None)

        result = None
        async with self.redisConnection.conn.pipeline(transaction=True) as pipe:
            jobs_tree = await self.addNode(flow, {"parentOpts": parent_opts},opts.get("queuesOptions"), pipe)
            await pipe.execute()
            result = jobs_tree

        return result

    async def addBulk(self, flows: list[dict]):
        result = None
        async with self.redisConnection.conn.pipeline(transaction=True) as pipe:
            job_trees = await self.addNodes(flows, pipe)
            await pipe.execute()
            result = job_trees

        return result

    def close(self):
        """
        Close the flow instance.
        """
        return self.redisConnection.close()




from typing import TypedDict, Any
import redis.asyncio as redis


class WorkerOptions(TypedDict, total=False):
    autorun: bool
    """
    Condition to start processor at instance creation

    @default true
    """

    concurrency: int
    """
    Amount of jobs that a single worker is allowed to work on
    in parallel.

    @default 1
    @see https://docs.bullmq.io/guide/workers/concurrency
    """

    maxStalledCount: int
    """
    Amount of times a job can be recovered from a stalled state
    to the `wait` state. If this is exceeded, the job is moved
    to `failed`.

    @default 1
    """

    stalledInterval: int
    """
    Number of milliseconds between stallness checks.

    @default 30000
    """

    lockDuration: int
    """
    Duration of the lock for the job in milliseconds. The lock represents that
    a worker is processing the job. If the lock is lost, the job will be eventually
    be picked up by the stalled checker and move back to wait so that another worker
    can process it again.

    @default 30000
    """

    prefix: str
    """
    Prefix for all queue keys.
    """

    connection: dict[str, Any] | redis.Redis
    """
    Options for connecting to a Redis instance.
    """



from typing import TypedDict


class BackoffOptions(TypedDict, total=False):
    type: str | dict
    """
    Name of the backoff strategy.
    """

    delay: int
    """
    Delay in milliseconds.
    """



from typing import TypedDict
from bullmq.types import BackoffOptions, KeepJobs


class JobOptions(TypedDict, total=False):
    backoff: int | BackoffOptions
    """
    Backoff setting for automatic retries if the job fails.
    """
    
    jobId: str
    """
    Override the job ID - by default, the job ID is a unique
    integer, but you can use this setting to override it.

    If you use this option, it is up to you to ensure the
    jobId is unique. If you attempt to add a job with an id that
    already exists, it will not be added.
    """

    timestamp: int
    """
    Timestamp when the job was created.

    @defaultValue round(time.time() * 1000)
    """

    delay: int
    """
    An amount of milliseconds to wait until this job can be processed.
    Note that for accurate delays, worker and producers
    should have their clocks synchronized.

    @defaultValue 0
    """

    attempts: int
    """
    The total number of attempts to try the job until it completes.

    @defaultValue 0
    """

    removeOnComplete: bool | int | KeepJobs
    """
    If true, removes the job when it successfully completes
    When given a number, it specifies the maximum amount of
    jobs to keep, or you can provide an object specifying max
    age and/or count to keep. It overrides whatever setting is used in the worker.
    Default behavior is to keep the job in the completed set.
    """

    removeOnFail: bool | int | KeepJobs
    """
    If true, removes the job when it fails after all attempts.
    When given a number, it specifies the maximum amount of
    jobs to keep, or you can provide an object specifying max
    age and/or count to keep. It overrides whatever setting is used in the worker.
    Default behavior is to keep the job in the failed set.
    """

    stackTraceLimit: int
    """
    Limits the amount of stack trace lines that will be recorded in the stacktrace.
    """
